import importlib.resources

from loguru import logger


def get_non_templated_file(templated_file: str) -> str:
    """Generates a non-templated txt file, which basically only appending the text from the file

    Args:
        templated_file (str): The name of the templated file located in the assets

    Returns:
        str: The contents of the non-templated text file
    """
    logger.debug(f"Generate '{templated_file}'")
    with importlib.resources.open_text("turludock.assets.dockerfile_templates", templated_file) as f:
        str_output = f.read()
    str_output += "\n\n"
    return str_output


def generate_entrypoint() -> str:
    """Get entrypoint.txt as a string

    Returns:
        str: The entrypoint.txt as a string
    """
    return get_non_templated_file("entrypoint.txt")


def generate_cmd() -> str:
    """Get cmd.txt as a string

    Returns:
        str: The cmd.txt as a string
    """
    return get_non_templated_file("cmd.txt")


def generate_common_env_config() -> str:
    """Get common_env_config.txt as a string

    Returns:
        str: The common_env_config.txt as a string
    """
    return get_non_templated_file("common_env_config.txt")


def generate_install_common_packages() -> str:
    """Get install_common_packages.txt as a string

    Returns:
        str: The install_common_packages.txt as a string
    """
    return get_non_templated_file("install_common_packages.txt")


def generate_locale() -> str:
    """Get locale.txt as a string

    Returns:
        str: The locale.txt as a string
    """
    return get_non_templated_file("locale.txt")


def generate_ohmyzsh() -> str:
    """Get oh_my_zsh.txt as a string

    Returns:
        str: The oh_my_zsh.txt as a string
    """
    return get_non_templated_file("oh_my_zsh.txt")


def generate_terminator() -> str:
    """Get terminator.txt as a string

    Returns:
        str: The terminator.txt as a string
    """
    return get_non_templated_file("terminator.txt")


def generate_conan() -> str:
    """Get conan.txt as a string

    Returns:
        str: The conan.txt as a string
    """
    return get_non_templated_file("conan.txt")


def generate_vscode() -> str:
    """Get vscode.txt as a string

    Returns:
        str: The vscode.txt as a string
    """
    return get_non_templated_file("vscode.txt")


def generate_meld() -> str:
    """Get meld.txt as a string

    Returns:
        str: The meld.txt as a string
    """
    return get_non_templated_file("meld.txt")


def generate_cpplint() -> str:
    """Get cpplint.txt as a string

    Returns:
        str: The cpplint.txt as a string
    """
    return get_non_templated_file("cpplint.txt")


def generate_mesa(use_latest: bool = True) -> str:
    """Get the mesa install command as a string

    Args:
        use_latest (bool, optional): Whether to use the latest mesa version. Defaults to True.

    Returns:
        str: The mesa Dockerfile instructions as a string
    """
    if use_latest:
        filename = "mesa_latest.txt"
    else:
        filename = "mesa.txt"
    return get_non_templated_file(filename)



import importlib.resources
from string import Template
from typing import Any, Dict, List

from loguru import logger

from turludock.config_sanity import (
    check_if_cmake_version_exists,
    check_if_llvm_version_exists,
    check_if_tmux_version_exists,
)
from turludock.helper_functions import (
    get_cpu_count_for_build,
    get_ros_major_version,
    get_ubuntu_version,
    is_ros_version_supported,
    is_version_greater,
    is_version_lower,
)


def populate_templated_file(mapping: Dict[str, str], templated_file: str) -> str:
    """Populates a templated file with the provided mapping and returns it as generated string

    Args:
        mapping (Dict[str, str]): The mapping of template variables to their values.
        templated_file (str): The name of the templated file.

    Returns:
        str: The populated templated file.
    """
    with importlib.resources.open_text("turludock.assets.dockerfile_templates", templated_file) as f:
        src = Template(f.read())
        str_output = src.substitute(mapping)
    str_output += "\n\n"
    return str_output


def _get_ubuntu_base_image(version: str, nvidia: bool) -> str:
    """Get the base image for the Docker build. It is used in "FROM <base_image>".

    Args:
        version (str): The semantic version of Ubuntu
        nvidia (bool): Whether the NVIDIA GPU driver is being used

    Returns:
        str: The Docker image name to use as the base image.
    """
    if nvidia:
        if is_version_lower(version, "16.04"):
            raise ValueError(f"Ubuntu version lower than 16.04 is not supported. You provided: {version}")

        # TODO(ATA): Not sure nvidia/opengl is really needed. Can we not just use 'ubuntu' with nvidia-docker-v2?
        return f"nvidia/opengl:1.2-glvnd-runtime-ubuntu{version}"
    else:
        return f"ubuntu:{version}"


def _get_base_image(yaml_config: Dict[str, Any]) -> str:
    """Return the supported base image name.

    This is used in the "FROM " part of the Dockerfile. See "from.txt" template.

    Args:
        yaml_config (dict): The configuration of the image to be built.

    Returns:
        str: The base image name.
    """
    ubuntu_version = get_ubuntu_version(yaml_config["ros_version"])
    if yaml_config["gpu_driver"] == "nvidia":
        uses_nvidia = True
        # Temp fix until nvidia releases nvidia/opengl for Ubuntu 24.04
        if is_version_greater(ubuntu_version["semantic"], "23.04"):
            uses_nvidia = False
            logger.warning(
                "'nvidia/opengl:1.2-glvnd-runtime' does not exist yet for Ubuntu 24.04. "
                + "Using 'ubuntu:20.04' as base image instead. This is experimental. "
                + "Please report any issues faced."
            )
    else:
        uses_nvidia = False
    return _get_ubuntu_base_image(ubuntu_version["semantic"], uses_nvidia)


def generate_from(yaml_config: Dict[str, Any]) -> str:
    """Generates the 'from.txt' templated file.

    Args:
        from_str (str): The value to be used for the 'From' template variable.

    Returns:
        yaml_config (dict): The image configuration in yaml format.
    """
    # Map the template variables
    base_image = _get_base_image(yaml_config)
    mapping = {"from": base_image}

    # Pick template based on ubuntu version
    ubuntu_version = get_ubuntu_version(yaml_config["ros_version"])
    if is_version_greater(ubuntu_version["semantic"], "23.04"):
        logger.debug(f"Generate 'from_2404.txt'. Input: {base_image}")
        return populate_templated_file(mapping, "from_2404.txt")
    else:
        logger.debug(f"Generate 'from.txt'. Input: {base_image}")
        return populate_templated_file(mapping, "from.txt")


def generate_header_info(docker_label_description: str, ros_version_short: str) -> str:
    """Generates the 'header_info.txt' templated file.

    Args:
        docker_label_description (str): The text for the "LABEL Description =" field
        ros_version_short (str): The ROS codename used in "LABEL com.turlucode.ros.version="

    Returns:
        str: The populated 'header_info.txt' file as a string.

    Raises:
        ValueError: If the provided ROS version is not supported.
    """
    logger.debug(f"Generate 'header_info.txt'. Input: {docker_label_description}, {ros_version_short}")

    # Check if provided version is supported
    if not is_ros_version_supported(ros_version_short):
        raise ValueError(f"ROS version '{ros_version_short}' not supported. Check your configuration.")

    # Map the template variables
    mapping = {
        "docker_label_description": docker_label_description,
        "ros_version_short": ros_version_short,
    }

    # Populate the templated file
    return populate_templated_file(mapping, "header_info.txt")


def generate_cmake(version: str) -> str:
    """Generates the 'cmake.txt' templated file.

    Args:
        version (str): The version of CMake to be used.

    Returns:
        str: The populated 'cmake.txt' file as a string.
    """
    logger.debug(f"Generate 'cmake.txt'. Input: {version}")

    # Check if provided version exists in remote
    check_if_cmake_version_exists(version)

    # Map the template variables
    mapping = {"cmake_version": version, "num_of_cpu": get_cpu_count_for_build()}

    # Populate the templated file
    return populate_templated_file(mapping, "cmake.txt")


def generate_tmux(version: str) -> str:
    """Generates the 'tmux.txt' templated file.

    Args:
        version (str): The version of tmux to be used.

    Returns:
        str: The populated 'tmux.txt' file as a string.
    """
    logger.debug(f"Generate 'tmux.txt'. Input: {version}'")

    # Check if provided version exists in remote
    check_if_tmux_version_exists(version)

    # Map the template variables
    mapping = {"tmux_version": version, "num_of_cpu": get_cpu_count_for_build()}

    # Populate the templated file
    return populate_templated_file(mapping, "tmux.txt")


def generate_llvm(version: str) -> str:
    """Generates the 'llvm.txt' templated file.

    Args:
        version (str): The version of LLVM to be used.

    Returns:
        str: The populated 'llvm.txt' file as a string.
    """
    logger.debug(f"Generate 'llvm.txt'. Input: {version}")

    # Check if provided version is supported
    check_if_llvm_version_exists(version)

    # Map the template variables
    mapping = {"llvm_version": version}

    # Populate the templated file
    return populate_templated_file(mapping, "llvm.txt")


def generate_ros(ros_version_codename: str) -> str:
    """Generates the 'ros1.txt' or 'ros2.txt' templated file, which is responsible for installing ROS 1 or 2

    The selection of ROS 1 or 2 is based on the provided ROS codename.

    Args:
        ros_version_codename (str): The ROS version as codename.

    Returns:
        str: The populated 'ros1.txt' file as a string.

    Raises:
        ValueError: If the provided ROS version is not supported.
    """
    # Determine if it ROS1 or ROS2
    if get_ros_major_version(ros_version_codename) == 1:
        template_file = "ros1.txt"
    else:
        template_file = "ros2.txt"

    logger.debug(f"Generate '{template_file}'. Input: {ros_version_codename}")

    # Check if provided version is supported
    if not is_ros_version_supported(ros_version_codename):
        raise ValueError("ROS version not supported. Check your configuration.")

    # Map the template variables
    mapping = {"ros_version_short": ros_version_codename}

    # Populate the templated file
    return populate_templated_file(mapping, template_file)


def generate_extra_packages_label(extra_packages_list: List[str]) -> str:
    """Generates the 'extra_packages_label.txt' templated file

    This file is used to add extra packages to the image.

    Args:
        extra_packages_list (List[str]): The list of extra packages to be installed in the docker image

    Returns:
        str: The populated 'extra_packages_label.txt' file as a string.
    """
    logger.debug(f"Generate 'extra_packages_label.txt'. Input: {extra_packages_list}")

    # Map the template variables
    list_as_str = " ".join(extra_packages_list)
    mapping = {"extra_packages_list": list_as_str}

    # Populate the templated file
    return populate_templated_file(mapping, "extra_packages_label.txt")



import argparse
import os
from typing import Tuple

from loguru import logger

from turludock.helper_functions import get_program_version


class PrintVersionAction(argparse.Action):
    """Action class to print the version of the program."""

    def __call__(self, parser, namespace, values, option_string=None):
        """The actual call to the action."""
        logger.info(f"turludock v{get_program_version()}")
        parser.exit()


def print_command_help(args: argparse.Namespace, parser: dict) -> None:
    """Print help message for the command specified by the user.

    Args:
        args (argparse.Namespace): The parsed command line arguments.
        parser (dict): Dictionary containing the command line arguments parsers.
    """
    if args.command == "build":
        parser["build"].print_help()
    elif args.command == "generate":
        parser["gen"].print_help()
    elif args.command == "which":
        if args.which is None:
            parser["which"].print_help()
        else:
            if args.which == "presets":
                parser["wpre"].print_help()
            elif args.which == "ros":
                parser["wros"].print_help()
            elif args.which == "cuda":
                parser["wnv"].print_help()
    else:
        parser.print_help()


def check_arguments(args: argparse.Namespace) -> None:
    """Check if the command line arguments are valid.

    Checks the arguments depending on the command.

    Args:
        args (argparse.Namespace): The parsed command line arguments.

    Raises:
        ValueError: If the arguments are invalid.
    """
    if args.command == "build":
        if args.e and args.c:
            raise ValueError("Provide either argument '-c' or argument '-e'\n")
        if not args.e and not args.c:
            raise ValueError("Provide either argument '-c' or argument '-e'\n")
    elif args.command == "generate":
        if args.e and args.c:
            raise ValueError("Provide either argument '-c' or argument '-e'\n")
        if not args.e and not args.c:
            raise ValueError("Provide either argument '-c' or argument '-e'\n")
        if not args.path:
            raise ValueError("The following arguments are required: path\n")
        if not os.path.isdir(args.path):
            raise ValueError(f"The path '{args.path}' is not a valid directory.\n")
    elif args.command == "which":
        if args.which is None:
            raise ValueError("You need to provide one of the following sub-commands: presets, ros, cuda\n")
        if args.which == "presets":
            pass
        elif args.which == "ros":
            pass
        elif args.which == "cuda":
            pass
        else:
            raise ValueError("Only the following sub-commands are supported: presets, ros, cuda\n")
    else:
        raise ValueError(f"Unknown command '{args.command}'.\n")


def parse_command_line_args() -> Tuple[argparse.Namespace, bool]:
    """Parse the command line arguments.

    Returns:
        Tuple[argparse.Namespace, bool]: A tuple containing the parsed command line arguments as
        an argparse.Namespace object, and a boolean indicating whether the arguments are valid or not.
    """
    parser = dict()
    parser["main"] = argparse.ArgumentParser(
        description="ROS docker container generator " + f"| turludock v{get_program_version()} | TurluCode"
    )
    parser["main"].add_argument(
        "--version", action=PrintVersionAction, nargs=0, help="Show the program version and exit"
    )

    subparsers = parser["main"].add_subparsers(title="Commands", description="", help="additional help", dest="command")
    subparsers.required = True  # Ensure that a sub-command is required

    # Sub-command 'build'
    parser["build"] = subparsers.add_parser("build", help="Generates and also builds the image using 'docker build'")
    parser["build"].add_argument(
        "-c", type=str, metavar="YAML_CONFIG", help="Provide the Dockerfile configuration .yaml file"
    )
    parser["build"].add_argument(
        "-e",
        type=str,
        metavar="CONFIG_NAME",
        help='Choose an existing pre-configuration. Check with "turludock which presets"',
    )
    parser["build"].add_argument(
        "--tag", type=str, metavar="TAG", help='Name and optionally a tag (format: "name:tag")'
    )
    parser["build"].add_argument(
        "--no-cache", action="store_true", default=False, help="Do not use cache when building the image"
    )
    parser["build"].add_argument(
        "-v", "--verbose", action="store_true", default=False, help="Shows the complete docker build output"
    )
    parser["build"].add_argument("-d", "--debug", action="store_true", default=False, help="Enable debug mode")

    # Sub-command 'generate'
    parser["gen"] = subparsers.add_parser(
        "generate",
        help="Generates the Dockerfile and the required assets for manual build. " "Overwrites existing files!",
    )
    parser["gen"].add_argument(
        "-c", type=str, metavar="YAML_CONFIG", help="Provide the Dockerfile configuration .yaml file"
    )
    parser["gen"].add_argument(
        "-e",
        type=str,
        metavar="CONFIG_NAME",
        help='Choose an existing pre-configuration. Check with "turludock which presets"',
    )
    parser["gen"].add_argument(
        "path",
        type=str,
        help="The directory path where the Dockerfile and its assets should be generated. "
        "Contents will be overwritten!",
    )
    parser["gen"].add_argument("-d", "--debug", action="store_true", default=False, help="Enable debug mode")

    # Sub-command 'which'
    parser["which"] = subparsers.add_parser("which", help="List available pre-configurations for generating ROS images")

    which_parsers = parser["which"].add_subparsers(dest="which")

    # Sub-command 'which presets'
    parser["wpre"] = which_parsers.add_parser(
        "presets", help="List available pre-configurations for directly generating ROS images"
    )
    parser["wpre"].add_argument("-d", "--debug", action="store_true", default=False, help="Enable debug mode")
    # Sub-command 'which ros'
    parser["wros"] = which_parsers.add_parser("ros", help="List supported ROS versions")
    parser["wros"].add_argument("-d", "--debug", action="store_true", default=False, help="Enable debug mode")
    # Sub-command 'which cuda'
    parser["wnv"] = which_parsers.add_parser("cuda", help="List supported CUDA/cuDNN versions for a given ROS version")
    parser["wnv"].add_argument("ros_codename", type=str, help="ROS codename string, e.g. noetic")
    parser["wnv"].add_argument("-d", "--debug", action="store_true", default=False, help="Enable debug mode")

    # Parse arguments
    args = parser["main"].parse_args()

    # Check them before returning
    try:
        check_arguments(args)
        return args, True
    except ValueError as e:
        logger.error(f"{e}")
        print_command_help(args, parser)
        return args, False



# Supported ROS codenames mapped to either ROS 1 or ROS 2
ROS_VERSION_MAP = {"noetic": 1, "humble": 2, "iron": 2, "jazzy": 2}

# ROS release dates since epoch
ROS_VERSION_RELEASE_DATE_MAP = {"noetic": 1590264000, "humble": 1653336000, "iron": 1684872000, "jazzy": 1716494400}



import importlib.metadata
import importlib.resources
import multiprocessing
import os
import re
import subprocess
from typing import List

import requests
import urllib3
from loguru import logger
from packaging.version import InvalidVersion, Version

import turludock.constants as constants


def get_module_name() -> str:
    """Returns the name of the module.

    Returns:
        str: The name of the module
    """
    # TODO(ATA): Maybe use: from utils import get_module_name
    return "turludock"


def check_if_remote_tag_exists(remote_url: str, tag_name: str) -> bool:
    """Check if a specific tag exists in a remote repository.

    Args:
        remote_url (str): The URL of the remote repository.
        tag_name (str): The name of the tag to check.

    Returns:
        bool: True if the tag exists, False otherwise.

    Raises:
        Exception: If a general exception occurs while checking the tag.
    """
    try:
        # Run the git command to list remote tags
        result = subprocess.run(
            ["git", "ls-remote", "--tags", remote_url],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )

        # Check if the command was successful
        if result.returncode != 0:
            logger.error(f"Git command not successful: {result.stderr}")
            return False

        # Check if the tag exists in the remote repository
        tags = result.stdout
        if f"refs/tags/{tag_name}" in tags:
            logger.debug(f"Tag '{tag_name}' exists in remote")
            return True
        else:
            logger.error(f"Tag {tag_name} not found in remote! Check provided tag!")
            return False
    except Exception as e:
        logger.error(f"Could not check if remote-tag in remote exists: {e}")
        return False


def get_github_latest_version_tag(owner: str, repo: str) -> str:
    """Fetches the latest version tag from a GitHub repository.

    Args:
        owner (str): The owner of the repository.
        repo (str): The name of the repository.

    Raises:
        ValueError: In case we get stats 200 from github API - unsuccessful request
        ValueError: Or if there are no valid versions found

    Returns:
        str: The latest version tag or None if no valid tags are found.
    """
    logger.debug(f"Trying to get latest version-tag from github for {owner}/{repo}...")

    # GitHub API URL for fetching tags of the repository
    url = f"https://api.github.com/repos/{owner}/{repo}/tags"
    response = requests.get(url, timeout=10)

    # Check if the request was successful
    if response.status_code != 200:
        raise ValueError(f"Cannot determine tag-version: Error fetching tags: {response.status_code}")

    # Parse the JSON response to get the list of tags
    tags = response.json()

    # Iterate over each tag to check if it follows semantic versioning
    valid_versions = []
    for tag in tags:
        tag_name = tag["name"]
        try:
            version = Version(tag_name.lstrip("v"))  # Remove 'v' prefix if present
            valid_versions.append((version, tag_name))
        except InvalidVersion:
            continue

    # Check if there are any valid versions found
    if not valid_versions:
        raise ValueError(f"Cannot determine tag-version: No valid versions found for {owner}/{repo} on github.")

    latest_version = max(valid_versions, key=lambda v: v[0])
    logger.debug(f"Latest version found for {owner}/{repo} is {latest_version[1]}")
    return latest_version[1]


def get_llvm_supported_versions() -> List[int]:
    """Fetches a list of supported LLVM versions from the official LLVM APT repository.

    The function fetches the content of the URL ``https://apt.llvm.org/llvm.sh`` and
    uses a regular expression to extract the supported LLVM version numbers from the
    content.

    Returns:
        List[int]: A list of integers representing the supported LLVM version numbers
    """
    # Get LLVM install script which contains info about the supported versions
    url = "https://apt.llvm.org/llvm.sh"
    try:
        # fixes warning: InsecureRequestWarning: Unverified HTTPS request is being made to host 'apt.llvm.org'
        # when using verify=False in requests.get()
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        # Get the content of the URL
        response = requests.get(url, timeout=10, verify=False)
    except Exception as e:
        logger.error(f"Could not get supported LLVM versions. Requests.get() error: {e}")
        raise

    # Check if the request was successful
    if response.status_code != 200:
        response.raise_for_status()

    # Define the regex pattern to match lines with LLVM_VERSION_PATTERNS[version]="-(version)"
    # the ="-(version)" is important so we can capture the group and extract the version integer
    pattern = re.compile(r'LLVM_VERSION_PATTERNS\[\d+\]="-(\d+)"')

    # Split content into lines and search for the pattern
    supported_versions = list()
    lines = response.text.splitlines()
    for line in lines:
        match = pattern.search(line)
        if match:
            version_number_int = int(match.group(1))  # Extract the number - this is still a int
            supported_versions.append(version_number_int)

    if len(supported_versions) == 0:
        raise ValueError(
            "No supported llvm could be extracted from https://apt.llvm.org/llvm.sh. "
            "Versioning might have changed, contact developers of this tool."
        )

    return supported_versions


def get_llvm_latest_version() -> int:
    """Fetches the latest supported LLVM version number.

    The function returns the maximum of the supported LLVM version numbers as returned by
    'get_llvm_supported_versions'.

    Returns:
        int: The latest supported LLVM version number
    """
    return max(get_llvm_supported_versions())


def list_yaml_files(directory: str) -> List[str]:
    """Walk through a directory and returns a list of all YAML files.

    Args:
        directory (str): The directory to start searching from.

    Returns:
        List[str]: A list of paths to YAML files.
    """
    yaml_files = []
    for root, _, files in os.walk(directory):
        for file in files:
            if file.endswith(".yaml"):
                # file_dict = {'filename': file,
                #              'full_path': os.path.join(root, file)}
                # yaml_files.append(file_dict)
                yaml_files.append(os.path.join(root, file))
    return yaml_files


def list_packaged_yaml_files(asset_folder: str) -> List[str]:
    """List all YAML files from a package.

    Args:
        asset_folder (str): The folder name of where the YAML files are located.

    Returns:
        List[str]: A list of paths to YAML files inside the package
    """
    # ref = importlib.resources.files(package)
    ref = importlib.resources.files(get_module_name())
    with importlib.resources.as_file(ref) as resource_path:
        try:
            asset_path = os.path.join(resource_path, asset_folder)
            return list_yaml_files(asset_path)
        except Exception as e:
            logger.error(f"Could not list yaml files in asset folder {asset_folder}: {e}")
            raise


def get_cpu_count_for_build() -> int:
    """Get the number of CPUs available for a build.

    These are usually being used by make, e.g. make -j<num_of_cpu>
    The idea is to help speed up the build process of the Dockerfile.

    Returns:
        int: The number of CPUs available for the build.
    """
    no_of_cpu = multiprocessing.cpu_count()
    no_of_cpu_for_build = no_of_cpu - 2

    if no_of_cpu_for_build <= 0:
        no_of_cpu_for_build = 1

    return no_of_cpu_for_build


def get_ubuntu_version(ros_version_codename: str) -> dict:
    """Creates a mapper that maps the ROS codename to the Ubuntu version

    The ROS codename is assumed to be checked already with the config_sanity.py

    Args:
        ros_version_codename (str): The ROS distribution codename

    Returns:
        dict: The corresponding Ubuntu version
    """
    if is_ros_version_supported(ros_version_codename):
        version_map = {
            "noetic": {"flat": "ubuntu2004", "semantic": "20.04"},
            "humble": {"flat": "ubuntu2204", "semantic": "22.04"},
            "iron": {"flat": "ubuntu2204", "semantic": "22.04"},
            "jazzy": {"flat": "ubuntu2404", "semantic": "24.04"},
        }
        return version_map[ros_version_codename]
    else:
        raise ValueError(f"Error mapping ROS '{ros_version_codename}' to Ubuntu version.")


def get_ros_major_version(ros_version_codename: str) -> int:
    """Creates a mapper that maps the ROS codename to the ROS major version

    The ROS codename is assumed to be checked already with the config_sanity.py

    Args:
        ros_version_codename (str): The ROS distribution codename

    Returns:
        int: The corresponding major ROS version, i.e. 1 or 2
    """
    version_map = constants.ROS_VERSION_MAP

    return version_map[ros_version_codename]


def is_ros_version_supported(ros_version_codename: str) -> bool:
    """Checks if a given ROS version is supported.

    Args:
        ros_version_codename (str): The codename of the ROS distribution.

    Returns:
        bool: True if the ROS version is supported, False otherwise.
    """
    try:
        ros_major_version = get_ros_major_version(ros_version_codename)
        logger.debug(f"ROS {ros_major_version} '{ros_version_codename}' version supported.")
        return True
    except KeyError:
        logger.error(f"Unsupported ROS version: {ros_version_codename}")
        return False


def is_version_lower(version_to_check: str, reference_version: str) -> bool:
    """Checks if a given version is lower than a reference version.

    Args:
        version_to_check (str): The version to check.
        reference_version (str): The reference version.

    Returns:
        bool: True if the version is lower than the reference version, False otherwise.
    """
    version = Version(version_to_check)
    reference = Version(reference_version)
    return version < reference


def is_version_greater(version_to_check: str, reference_version: str) -> bool:
    """Checks if a given version is greater than a reference version.

    Args:
        version_to_check (str): The version to check.
        reference_version (str): The reference version.

    Returns:
        bool: True if the version is greater than the reference version, False otherwise.
    """
    version = Version(version_to_check)
    reference = Version(reference_version)
    return version > reference


def get_program_version() -> str:
    """Gets the program version.

    See pyproject.toml

    Returns:
        str: The version
    """
    return importlib.metadata.version("turludock")



from typing import Any, Dict, Optional, Union

from loguru import logger

from turludock.config_parser import print_configuration
from turludock.generate_non_templated_files import (
    generate_cmd,
    generate_common_env_config,
    generate_conan,
    generate_cpplint,
    generate_entrypoint,
    generate_install_common_packages,
    generate_locale,
    generate_meld,
    generate_mesa,
    generate_ohmyzsh,
    generate_terminator,
    generate_vscode,
)
from turludock.generate_nvidia_templated_files import (
    generate_cuda_base,
    generate_cuda_devel,
    generate_cuda_runtime,
    generate_cudnn_devel,
    generate_cudnn_runtime,
)
from turludock.generate_templated_files import (
    generate_cmake,
    generate_extra_packages_label,
    generate_from,
    generate_header_info,
    generate_llvm,
    generate_ros,
    generate_tmux,
)
from turludock.helper_functions import (
    get_github_latest_version_tag,
    get_llvm_latest_version,
    get_ros_major_version,
    get_ubuntu_version,
    is_version_greater,
    is_version_lower,
)


def _get_item_from_extra_packages(extra_packages: list, package_name: str) -> Optional[Union[dict, str]]:
    """Get the dict or the str of a given package name

    Since the yaml list can contain either dictionaries or strings, here were trying to find the dictionary
    or the string that matches the given package_name.

    If a dict is found, return that dict. If a str is found, return that str. Otherwise return None.

    Args:
        extra_packages (List[Union[dict, str]]): A list of extra packages which could be either
            dictionaries or str.
        package_name (str): The name of the package to search for.

    Returns:
        Optional[Union[dict, str]]: If a dict which matches the package_name was found, return that dict. If a str
            which matches the package_name was found, return that str. Otherwise return None.
    """
    for item in extra_packages:
        if isinstance(item, dict) and package_name in item:
            return item
        if isinstance(item, str) and item == package_name:
            return item
    return None


def _generate_description(yaml_config: Dict[str, Any]) -> str:
    """Generate a description to be using in the "LABEL Description=" of the Dockerfile.

    See "header_info.txt" template

    Args:
        yaml_config (dict): The image configuration in yaml format.

    Returns:
        str: The generated description
    """
    ros_version = yaml_config["ros_version"]
    ros_version_str = f"ROS {get_ros_major_version(ros_version)} {ros_version.capitalize()}"

    # Ubuntu version
    ubuntu_version = get_ubuntu_version(ros_version)
    ubuntu_version_str = f"Ubuntu {ubuntu_version['semantic']}"

    # GPU driver
    gpu_driver_str = f"{yaml_config['gpu_driver']}"

    return f"{ros_version_str} | {ubuntu_version_str} | {gpu_driver_str}"


def _get_package_version(yaml_config: dict, package_name: str) -> str:
    """Get the version of a package from the yaml config or get the latest version by looking usually online.

    If the package is configured explicitly in the 'extra_packages' list, use the version
    specified there. Otherwise, infer that the latest version should be used.

    Args:
        yaml_config (dict): The parsed yaml config
        package_name (str): The name of the package to get the version of

    Returns:
        str: The version of the package

    Raises:
        ValueError: If the package is not known or the latest version cannot be determined
    """
    # Check if we even have key 'extra_packages'
    if "extra_packages" in yaml_config:
        item = _get_item_from_extra_packages(yaml_config["extra_packages"], package_name)
        if isinstance(item, dict):
            get_latest = False
        else:
            get_latest = True
    # If package not configured explicitly, always pick the latest version
    # This is only applicable for the default packages and never happens to the extra-packages
    else:
        get_latest = True

    # Infer and get the 'latest' version
    if get_latest:
        if package_name == "cmake":
            return get_github_latest_version_tag("Kitware", "CMake")
        elif package_name == "tmux":
            return get_github_latest_version_tag("tmux", "tmux")
        elif package_name == "llvm":
            return str(get_llvm_latest_version())
        else:
            raise ValueError(f"get_package_version() unknown package: {package_name}")
    # Get the version as defined in the .yaml
    else:
        return item[package_name]


def generate_dockerfile(yaml_config: Dict[str, Any]) -> str:
    """Generate the actual Dockerfile from a given yaml configuration.

    Based on YAML configuration we populate all templates to finally generate the Dockerfile.

    Args:
        yaml_config (dict): The image configuration in yaml format.

    Returns:
        str: The generated Dockerfile.
    """
    # Print configuration
    print("")
    logger.info("Configuration:")
    print_configuration(yaml_config)

    # Generate Dockerfile
    dockerfile = ""

    # Base image
    dockerfile += generate_from(yaml_config)

    # Meta data
    dockerfile += generate_header_info(_generate_description(yaml_config), yaml_config["ros_version"])

    # Common configuration for all images
    dockerfile += generate_common_env_config()
    dockerfile += generate_install_common_packages()
    dockerfile += generate_locale()
    dockerfile += generate_cmake(_get_package_version(yaml_config, "cmake"))
    dockerfile += generate_terminator()
    dockerfile += generate_ohmyzsh()

    # GPU driver
    ubuntu_version = get_ubuntu_version(yaml_config["ros_version"])
    if yaml_config["gpu_driver"] == "mesa":
        if is_version_lower(ubuntu_version["semantic"], "20.04"):
            logger.warning(
                "Using mesa driver. However latest mesa drivers cannot be installed for "
                + f"Ubuntu {ubuntu_version['semantic']}. This means if you use laster Intel/AMD GPU "
                + "you most likely not going to get HW acceleration and programs like RViZ will use "
                + "software rendering. For more info check ppa:kisak/kisak-mesa"
            )
            # use mesa from default ubuntu repos
            dockerfile += generate_mesa(False)
        elif is_version_greater(ubuntu_version["semantic"], "22.04"):
            # use mesa from default ubuntu repos - those should be the latest
            dockerfile += generate_mesa(False)
        else:
            # use latest mesa provided by custom ppa
            dockerfile += generate_mesa(True)
    else:
        # no need to install something for NVIDIA
        pass

    # Add CUDA/cuDNN
    ubuntu_version = get_ubuntu_version(yaml_config["ros_version"])
    if "cuda_version" in yaml_config:
        dockerfile += generate_cuda_base(yaml_config["cuda_version"], ubuntu_version["flat"])
        dockerfile += generate_cuda_devel(yaml_config["cuda_version"], ubuntu_version["flat"])
        dockerfile += generate_cuda_runtime(yaml_config["cuda_version"], ubuntu_version["flat"])
        if "cudnn_version" in yaml_config:
            dockerfile += generate_cudnn_devel(
                yaml_config["cuda_version"], yaml_config["cudnn_version"], ubuntu_version["flat"]
            )
            # TODO(ATA): Devel already contains what runtime has. So it redundant.
            # But is it the case for cuDNN implementation?
            dockerfile += generate_cudnn_runtime(
                yaml_config["cuda_version"], yaml_config["cudnn_version"], ubuntu_version["flat"]
            )

    # Add ROS
    dockerfile += generate_ros(yaml_config["ros_version"])

    # Add the extra-packages
    extra_packages_label_list = list()
    if "extra_packages" in yaml_config:
        for item in yaml_config["extra_packages"]:
            if isinstance(item, dict):
                package_name = next(iter(item), None)
            elif isinstance(item, str):
                package_name = item
            else:
                raise ValueError("Item in 'extra_packages' should be either a string or a dict.")

            if package_name == "tmux":
                dockerfile += generate_tmux(_get_package_version(yaml_config, package_name))
            if package_name == "llvm":
                dockerfile += generate_llvm(_get_package_version(yaml_config, package_name))
            if package_name == "meld":
                dockerfile += generate_meld()
            if package_name == "cpplint":
                dockerfile += generate_cpplint()
            if package_name == "conan":
                dockerfile += generate_conan()
            if package_name == "vscode":
                dockerfile += generate_vscode()

            extra_packages_label_list.append(package_name)
    else:
        extra_packages_label_list.append("")
        logger.debug("Warning: generate_dockerfile(): No extra packages have been configured.")
    dockerfile += generate_extra_packages_label(extra_packages_label_list)

    # Finally, add the entrypoint and cmd parts
    dockerfile += generate_entrypoint()
    dockerfile += generate_cmd()

    # print(dockerfile)
    return dockerfile



import os
from typing import Any, Dict

from termcolor import colored

import turludock.config_sanity as config_sanity
from turludock.helper_functions import get_ros_major_version, get_ubuntu_version


def get_config_filename(yaml_full_path: str) -> str:
    """Get the filename of the given YAML file.

    Args:
        yaml_full_path (str): The full path to the YAML file

    Returns:
        str: The filename of the given YAML file
    """
    return os.path.basename(yaml_full_path)


def get_config_name(yaml_full_path: str) -> str:
    """Infer the name of the given YAML file.

    For now this is the name of the file without its file extension

    Args:
        yaml_full_path (str): The full path to the YAML file

    Returns:
        str: The corresponding name of the YAML file
    """
    return get_config_filename(yaml_full_path).replace(".yaml", "")


def check_dockerfile_config(config: Dict[str, Any]) -> None:
    """Checks if the given Dockerfile configuration is valid.

    Checks the following:
        * The required fields are present.
        * The GPU driver is supported.
        * The NVIDIA configuration is valid.
        * The ROS version is supported.
        * The list of extra packages is valid.

    Args:
        config (dict): The Dockerfile configuration.
    """
    # First check if required fields are present
    config_sanity.check_required_fields(config)
    # Check GPU drivers
    config_sanity.check_supported_gpu_drivers(config)
    # Check NVIDIA configuration
    config_sanity.check_nvidia_config(config)
    # Check ROS version
    config_sanity.check_supported_ros_version(config)
    # Check the list of extra packages
    config_sanity.check_extra_packages(config)


def print_configuration(yaml_config: Dict[str, Any]) -> None:
    """Prints the given YAML configuration in a human readable format.

    Args:
        yaml_config (dict): The configuration file parsed as YAML.
    """
    # Config name
    config_name = get_config_name(yaml_config["filename"])
    config_name_str = colored(f"{config_name: <20}", "green", attrs=["bold"])

    # ROS version
    ros_version = yaml_config["ros_version"]
    ros_version_str = f"ROS {get_ros_major_version(ros_version)} {ros_version.capitalize()}"

    # Ubuntu version
    ubuntu_version = get_ubuntu_version(ros_version)
    ubuntu_version_str = f"Ubuntu {ubuntu_version['semantic']}"

    # GPU driver
    gpu_driver_str = "GPU: " + yaml_config["gpu_driver"]

    # CUDA/CUDNN
    if "cuda_version" in yaml_config:
        cuda_version_str = f"CUDA: {yaml_config['cuda_version']}"
    else:
        cuda_version_str = "CUDA: -"
    if "cudnn_version" in yaml_config:
        cudnn_version_str = f"cuDNN: {yaml_config['cudnn_version']}"
    else:
        cudnn_version_str = "cuDNN: -"

    # List of packages
    extra_packages_str = ""
    if "extra_packages" in yaml_config:
        extra_packages_str += "Extra packages:"
        for pkg in yaml_config["extra_packages"]:
            if isinstance(pkg, dict):
                package_name = next(iter(pkg), None)
                extra_packages_str += f" {package_name}-{pkg[package_name]}"
            elif isinstance(pkg, str):
                extra_packages_str += f" {pkg}"
    else:
        extra_packages_str = "No extra packages"

    print(
        f"{config_name_str} {ros_version_str: <13} | {ubuntu_version_str} | "
        + f"{gpu_driver_str: <11} | {cuda_version_str: <12} | {cudnn_version_str: <15} | {extra_packages_str}"
    )



import os

from turludock.config_parser import get_config_filename, get_config_name
from turludock.helper_functions import list_packaged_yaml_files
from turludock.yaml_load import load_default_image_configuration


def configuration_exists(config_name: str) -> bool:
    """Check if a configuration with the given name exists as asset in our module.

    See 'turludock/assets/default_image_configurations' for supported configurations.

    Args:
        config_name (str): The name of the configuration to check.

    Returns:
        bool: True if the configuration exists, False otherwise.
    """
    yaml_config_files = list_packaged_yaml_files(os.path.join("assets", "default_image_configurations"))
    for yaml_config in yaml_config_files:
        valid_config_name = get_config_name(yaml_config)
        if config_name == valid_config_name:
            return True
    return False


def get_yaml_config(config_name: str) -> dict:
    """Get YAML configuration file based on the given configuration name.

    Args:
        config_name (str): The name of the configuration to retrieve.

    Returns:
        dict: The configuration loaded from the YAML file.

    Raises:
        ValueError: If the provided configuration name does not exist.
    """
    yaml_config_files = list_packaged_yaml_files(os.path.join("assets", "default_image_configurations"))
    yaml_config = None
    for yaml_config_full_path in yaml_config_files:
        if config_name == get_config_name(yaml_config_full_path):
            yaml_filename = get_config_filename(yaml_config_full_path)
            yaml_config = load_default_image_configuration(yaml_filename)
            return yaml_config
    if yaml_config is None:
        raise ValueError(
            f"Provided pre-configuration '{config_name}' doesn't exist! "
            + "List available with 'turludock which preset'"
        )



import sys

from loguru import logger

_logger_id = 0


def configure_logger(debug: bool = False) -> None:
    """Configure the logger.

    Args:
        debug: Whether to set the logger level to DEBUG. Otherwise level is set to INFO.

    The logger is configured to print messages in the format: <level>{message}</level>
    """
    global _logger_id
    logger.remove(_logger_id)

    if debug:
        level = "DEBUG"
    else:
        level = "INFO"

    # format="<level>{time:HH:mm:ss.SS}</level> | <level>{level: <8}</level> | <level>{message}</level>", level=level)
    # logger.level("INFO", color="<green>")

    _logger_id = logger.add(sys.stdout, colorize=True, format="<level>{message}</level>", level=level)



import importlib.resources
import os
import shutil

from loguru import logger


def get_package_permissions(package: str, resource: str) -> int:
    """Get the linux file-permissions of a resource from a package.

    Args:
        package (str): The name of the package the resource belongs to.
        resource (str): The name of the resource to get the permissions of.

    Returns:
        int: The permissions of the resource.

    Raises:
        Exception: If an error occurs while getting the permissions of the resource.
    """
    try:
        with importlib.resources.path(package, resource) as resource_path:
            return os.stat(resource_path).st_mode
    except Exception as e:
        logger.error(f"Could not get permissions of package resource {package}/{resource}: {e}")
        raise


def copy_file(src: str, dst: str) -> None:
    """Copy a file from a source to a destination.

    Args:
        src (str): The source file to copy.
        dst (str): The destination file path.

    Raises:
        FileNotFoundError: If the source file does not exist.
        PermissionError: If the program does not have sufficient permissions to copy the source to the destination.
        Exception: Catch-all for any other unexpected errors.
    """
    try:
        shutil.copy(src, dst)
        logger.debug(f"File copied successfully from {src} to {dst}.")
    except FileNotFoundError:
        logger.error(f"File {src} not found.")
    except PermissionError:
        logger.error(f"Permission denied to copy {src} to {dst}.")
    except Exception as e:
        logger.error(f"File copy: An error occurred: {e}")


def get_filename_from_path(full_path: str) -> str:
    """Returns the filename from a given full path.

    Args:
        full_path (str): The full path.

    Returns:
        str: The filename.
    """
    return os.path.basename(full_path)


def copy_resource(package: str, resource_name: str, destination_path: str) -> None:
    """
    Copies a resource file from the specified module to the destination path.

    Args:
    - resource_name: The name of the resource file to copy.
    - destination_path: The file path where the resource should be copied.
    """
    ref = importlib.resources.files(package) / resource_name
    with importlib.resources.as_file(ref) as resource_path:
        try:
            shutil.copy(resource_path, destination_path)
            logger.debug(f"Successfully copied '{resource_path}' to '{destination_path}'.")
        except FileNotFoundError:
            logger.error(f"File {resource_name} not found.")
        except PermissionError:
            logger.error(f"Permission denied to copy {resource_path} to {destination_path}.")
        except Exception as e:
            logger.error(f"File copy: An error occurred: {e}")


def create_clean_directory(dir_path: str):
    """Creates a directory at the specified path, ensuring it is empty by deleting it first if it already exists.

    Parameters:
        dir_path (str): The path to the directory to be created or cleaned.

    Raises:
        PermissionError: If the program does not have sufficient permissions to delete or create directories.
        Exception: Catch-all for any other unexpected errors.
    """
    try:
        # Check if the directory exists
        if os.path.exists(dir_path):
            # If it exists, delete it
            shutil.rmtree(dir_path)
        # Create the directory
        os.makedirs(dir_path)
    except PermissionError as e:
        logger.error(f"Could not create a clean directory in {dir_path}. Permission error: {e}")
    except Exception as e:
        # Catch-all for any other exceptions
        logger.error(f"Could not create a clean directory in {dir_path}. An unexpected error occurred: {e}")


def to_absolute_path(path: str) -> str:
    """Checks if a path is absolute or relative. If it's relative, converts it to an absolute path.

    Parameters:
        path (str): The path to be checked and converted.

    Returns:
        str: The absolute path.
    """
    if os.path.isabs(path):
        # If path is already absolute, return it as it is
        return path
    else:
        # If path is relative, convert it to absolute
        return os.path.abspath(path)



import re
from typing import Optional, Tuple

from rich.progress import BarColumn, Progress, TextColumn, TimeElapsedColumn


class BuildProgress:
    """A class used to track progress of a docker build process."""

    def __init__(self) -> None:
        """Initializes a BuildProgress object.

        The object is used to track progress of a docker build process.

        The Progress object is configured to show the description of the task, a progress bar,
        the percentage completed and the time elapsed.

        The object has three attributes:
            - total_tasks: The total number of tasks
            - is_initialized: Whether the progress bar is initialized or not
            - progress: The Progress object
            - task: The Task object representing the current task. This is set when the progress bar
                      is initialized.
        """
        self.total_tasks = 0
        self.is_initialized = False
        self.progress = Progress(
            TextColumn("[bold blue]{task.description}"),
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.1f}%",
            TimeElapsedColumn(),
        )
        self.task = None

    def find_and_parse_extra_step(self, status_msg: str) -> Tuple[bool, Optional[int], Optional[int]]:
        """Parses the docker build output and looks for the "Step m/n" pattern.

        This is then used to determine the progress of the docker build command.

        The function uses regex to find the pattern in the input string.
        The function returns a tuple with a boolean indicating whether the pattern was found
        or not and two optional integers representing the "Step m/n".

        Args:
            status_msg (str): The string to search for the pattern in.

        Returns:
            Tuple[bool, Optional[int], Optional[int]]: A tuple with a boolean stating if parsing was successful
            and one integer returning the current step and another one returning the total steps.
            If parsing was unsuccessful, the tuple is (False, None, None).
        """
        # Define the regex pattern to match "Step NUMBER_ONE/NUMBER_TWO :"
        pattern = r"Step (\d+)/(\d+) :"

        # Use re.search to find the pattern in the input string
        match = re.search(pattern, status_msg)

        if match:
            # Extract the two numbers and convert them to integers
            current_step = int(match.group(1))
            total_steps = int(match.group(2))
            return True, current_step, total_steps
        else:
            return False, None, None

    def _start(self, total_tasks: int) -> None:
        """Initializes the progress bar based on the total steps the build command has.

        Args:
            total_tasks (int): The total number of steps the build command has.
        """
        self.total_tasks = total_tasks
        self.task = self.progress.add_task("[green]|Building...", total=self.total_tasks)
        self.progress.start()
        self.is_initialized = True

    def finish(self) -> None:
        """Stops the progress bar."""
        self.progress.stop()

    def advance(self, build_status_msg: str) -> None:
        """Advances the progress bar based on the docker build status message.

        To determine the progress, the function uses regex to find the "Step m/n" pattern.

        Args:
            build_status_msg (str): The docker build status message.
        """
        # Parse docker build status message to check progress
        found, _, total_tasks = self.find_and_parse_extra_step(build_status_msg)

        # If status message contains progress update the bar
        if found:
            if not self.is_initialized:
                self._start(total_tasks)
            # Update bar
            # print(f"Step: {step}/{total_tasks}")
            self.progress.advance(self.task, advance=1)



import importlib.resources
from typing import Any

import yaml
from loguru import logger


def load_yaml_file(file_path: str) -> dict:
    """
    Load and parse a YAML file.

    Parameters:
        file_path (str): The path to the YAML file.

    Returns:
        dict: The parsed YAML data as a dictionary.
    """
    try:
        with open(file_path, "r", encoding="utf-8") as yaml_file:
            yaml_data = yaml.safe_load(yaml_file)
            return yaml_data
    except IsADirectoryError as e:
        logger.error(f"Provided path '{file_path}' is not YAML file but a directory. Error: {e}")
        raise
    except FileNotFoundError as e:
        logger.error(f"Could not find YAML file '{file_path}'. Error: {e}")
        raise
    except yaml.YAMLError as e:
        logger.error(f"Error parsing YAML file '{file_path}'. Error: {e}")
        raise


def load_packaged_yaml(package: str, yaml_file: str) -> dict[str, Any]:
    """
    Load a YAML file packaged in a module.

    Parameters:
        package (str): The name of the module containing the YAML file.
        yaml_file (str): The name of the YAML file to load.

    Returns:
        dict[str, Any]: The parsed YAML data as a dictionary.
    """
    try:
        with importlib.resources.open_text(package, yaml_file) as f:
            return yaml.safe_load(f)
    except FileNotFoundError:
        logger.error(f"The file '{yaml_file}' was not found.")
        raise
    except yaml.YAMLError as exc:
        logger.error(f"Error parsing YAML file: {exc}")
        raise


def load_cuda_config() -> dict[str, Any]:
    """
    Load the YAML from our module that contains the supported CUDA configurations

    Returns:
        dict[str, Any]: The parsed YAML containing the supported CUDA configurations
    """
    return load_packaged_yaml("turludock.assets.config_files", "nvidia_cuda.yaml")


def load_cudnn_config() -> dict[str, Any]:
    """
    Load the YAML from our module that contains the supported cuDNN configurations

    Returns:
        dict[str, Any]: The parsed YAML containing the supported cuDNN configurations
    """
    return load_packaged_yaml("turludock.assets.config_files", "nvidia_cudnn.yaml")


def load_default_image_configuration(yaml_filename: str) -> dict[str, Any]:
    """
    Load the YAML from our module that contains the Dockerfile configuration

    Parameters:
        yaml_filename (str): The name of the YAML file configuration to load.

    Returns:
        dict[str, Any]: The parsed YAML configuration
    """
    yaml_config = load_packaged_yaml(
        package="turludock.assets.default_image_configurations",
        yaml_file=yaml_filename,
    )
    yaml_config.update({"filename": yaml_filename})
    return yaml_config



import sys

from loguru import logger

import turludock.generate_dockerfile_build_folder as generate_dockerfile_build_folder
from turludock.command_line_arguments_parser import parse_command_line_args
from turludock.docker_build import build_custom_image, build_pre_configured_image
from turludock.logger import configure_logger
from turludock.which_command import list_cuda_support, list_pre_configs, list_supported_ros_versions


def main() -> int:
    """
    Main entry point for the command line interface of the turludock tool.

    Returns:
        An integer indicating the exit status of the program. 0 indicates success,
        non-zero values indicate failure.
    """
    # Initialize the logger
    configure_logger()

    # Parse arguments
    args, parse_ok_ = parse_command_line_args()
    if not parse_ok_:
        sys.exit(1)

    # Enable debug mode
    if args.debug:
        configure_logger(True)

    # which
    if args.command == "which":
        if args.which == "presets":
            list_pre_configs()
        elif args.which == "ros":
            list_supported_ros_versions()
        elif args.which == "cuda":
            list_cuda_support(args.ros_codename)
        return 0
    # build
    if args.command == "build":
        # Build from pre-configuration
        try:
            # Build image from pre-configuration
            if args.e:
                build_args = {"tag": args.tag, "no_cache": args.no_cache, "verbose": args.verbose}
                build_pre_configured_image(args.e, build_args)
            # Build custom-image using user's .yaml config file
            elif args.c:
                build_args = {"tag": args.tag, "no_cache": args.no_cache, "verbose": args.verbose}
                build_custom_image(args.c, build_args)
        except Exception:
            logger.error("Error running 'build' command. Exit.")
            return 1
    # generate
    if args.command == "generate":
        try:
            # Generate from pre-configuration
            if args.e:
                generate_dockerfile_build_folder.generate_from_pre_config(args.e, args.path)
            # Generate using user's .yaml config file
            elif args.c:
                generate_dockerfile_build_folder.generate_from_user_config(args.c, args.path)
        except Exception:
            logger.error("Error running 'generate' command. Exit.")
            return 1


if __name__ == "__main__":
    main()



import importlib.resources
from string import Template

from loguru import logger

from turludock.config_sanity import is_cuda_cudnn_version_combination_supported, is_cuda_version_supported
from turludock.yaml_load import load_cuda_config, load_cudnn_config


def generate_cuda_base(cuda_version: str, ubuntu_version: str) -> str:
    """Populate the cuda_base.txt template file and return the generated string

    The CUDA and Ubuntu version are used to select the correct CUDA configuration.

    Args:
        cuda_version (str): CUDA version
        ubuntu_version (str): Ubuntu version

    Returns:
        str: dockerfile contents
    """
    logger.debug(f"Generate 'nvidia-cuda-base-{cuda_version}-{ubuntu_version}'")

    # Check if provided version is supported
    if not is_cuda_version_supported(cuda_version, ubuntu_version):
        raise ValueError("CUDA version not supported. Check your configuration.")

    cuda_config = load_cuda_config()[cuda_version][ubuntu_version]

    # Map the template variables
    mapping = {
        "cuda_version": cuda_version,
        "ubuntu_version": ubuntu_version,
        "nvidia_require_cuda": cuda_config["nvidia_require_cuda"],
        "cuda_cudart_version": cuda_config["cudart"],
        "cuda_compat_version": cuda_config["compat"],
    }

    # Populate the templated file
    package = "turludock.assets.dockerfile_templates.nvidia"
    with importlib.resources.open_text(package, "cuda_base.txt") as f:
        src = Template(f.read())
        str_output = src.substitute(mapping)
    str_output += "\n\n"
    return str_output


def generate_cuda_devel(cuda_version: str, ubuntu_version: str) -> str:
    """Populate the cuda_devel.txt template file and return the generated string

    The CUDA and Ubuntu version are used to select the correct CUDA configuration.

    Args:
        cuda_version (str): CUDA version
        ubuntu_version (str): Ubuntu version

    Returns:
        str: dockerfile contents
    """
    logger.debug(f"Generate 'nvidia-cuda-devel-{cuda_version}-{ubuntu_version}'")

    # Check if provided version is supported
    if not is_cuda_version_supported(cuda_version, ubuntu_version):
        raise ValueError("CUDA version not supported. Check your configuration.")

    cuda_config = load_cuda_config()[cuda_version][ubuntu_version]

    # Map the template variables
    mapping = {
        "cuda_cudart_version": cuda_config["cudart"],
        "cuda_lib_version": cuda_config["cuda_lib"],
        "cuda_nvml_version": cuda_config["nvml"],
        "cuda_nvprof_version": cuda_config["nvprof"],
        "cuda_libnpp_version": cuda_config["libnpp"],
        "cuda_libcusparse_version": cuda_config["libcusparse"],
        "cuda_libcublas_version": cuda_config["libcublas"],
        "cuda_libnccl_version": cuda_config["libnccl"],
        "cuda_nsight_compute_version": cuda_config["nsight_compute"],
    }

    # Populate the templated file
    package = "turludock.assets.dockerfile_templates.nvidia"
    with importlib.resources.open_text(package, "cuda_devel.txt") as f:
        src = Template(f.read())
        str_output = src.substitute(mapping)
    str_output += "\n\n"
    return str_output


def generate_cuda_runtime(cuda_version: str, ubuntu_version: str) -> str:
    """Populate the cuda_runtime.txt template file and return the generated string

    The CUDA and Ubuntu version are used to select the correct CUDA configuration.

    Args:
        cuda_version (str): CUDA version
        ubuntu_version (str): Ubuntu version

    Returns:
        str: dockerfile contents
    """
    logger.debug(f"Generate 'nvidia-cuda-runtime-{cuda_version}-{ubuntu_version}'")

    # Check if provided version is supported
    if not is_cuda_version_supported(cuda_version, ubuntu_version):
        raise ValueError("CUDA version not supported. Check your configuration.")

    cuda_config = load_cuda_config()[cuda_version][ubuntu_version]

    # Map the template variables
    mapping = {
        "cuda_lib_version": cuda_config["cuda_lib"],
        "cuda_libnpp_version": cuda_config["libnpp"],
        "cuda_nvtx_version": cuda_config["nvtx"],
        "cuda_libcusparse_version": cuda_config["libcusparse"],
        "cuda_libcublas_version": cuda_config["libcublas"],
        "cuda_libnccl_version": cuda_config["libnccl"],
    }

    # Populate the templated file
    package = "turludock.assets.dockerfile_templates.nvidia"
    with importlib.resources.open_text(package, "cuda_runtime.txt") as f:
        src = Template(f.read())
        str_output = src.substitute(mapping)
    str_output += "\n\n"
    return str_output


def generate_cudnn_devel(cuda_version: str, cudnn_version: str, ubuntu_version: str) -> str:
    """Populate the cudnn_devel.txt template file and return the generated string

    The CUDA, cuDNN and Ubuntu version are used to select the correct cuDNN configuration.

    Args:
        cuda_version (str): CUDA version
        cudnn_version (str): cuDNN version
        ubuntu_version (str): Ubuntu version

    Returns:
        str: dockerfile contents
    """
    logger.debug(f"Generate 'nvidia-cudnn-devel-{cudnn_version}-{ubuntu_version}'")

    # Check if provided version is supported
    if not is_cuda_cudnn_version_combination_supported(cuda_version, cudnn_version, ubuntu_version):
        raise ValueError("cuDNN version not supported. Check your configuration.")

    cudnn_config = load_cudnn_config()[cudnn_version][ubuntu_version]

    # Map the template variables
    mapping = {
        "libcudnn_package": cudnn_config["libcudnn_package"],
        "libcudnn_dev_package": cudnn_config["libcudnn_dev_package"],
        "libcudnn_version": cudnn_config["libcudnn_version"],
        "libcudnn_revision": cudnn_config["libcudnn_revision"],
    }

    # Populate the templated file
    package = "turludock.assets.dockerfile_templates.nvidia"
    with importlib.resources.open_text(package, "cudnn_devel.txt") as f:
        src = Template(f.read())
        str_output = src.substitute(mapping)
    str_output += "\n\n"
    return str_output


def generate_cudnn_runtime(cuda_version: str, cudnn_version: str, ubuntu_version: str) -> str:
    """Populate the cudnn_runtime.txt template file and return the generated string

    The CUDA, cuDNN and Ubuntu version are used to select the correct cuDNN configuration.

    Args:
        cuda_version (str): CUDA version
        cudnn_version (str): cuDNN version
        ubuntu_version (str): Ubuntu version

    Returns:
        str: dockerfile contents
    """
    logger.debug(f"Generate 'nvidia-cudnn-runtime-{cudnn_version}-{ubuntu_version}'")

    # Check if provided version is supported
    if not is_cuda_cudnn_version_combination_supported(cuda_version, cudnn_version, ubuntu_version):
        raise ValueError("cuDNN version not supported. Check your configuration.")

    cudnn_config = load_cudnn_config()[cudnn_version][ubuntu_version]

    # Map the template variables
    mapping = {
        "libcudnn_package": cudnn_config["libcudnn_package"],
        "libcudnn_version": cudnn_config["libcudnn_version"],
        "libcudnn_revision": cudnn_config["libcudnn_revision"],
    }

    # Populate the templated file
    package = "turludock.assets.dockerfile_templates.nvidia"
    with importlib.resources.open_text(package, "cudnn_runtime.txt") as f:
        src = Template(f.read())
        str_output = src.substitute(mapping)
    str_output += "\n\n"
    return str_output



from typing import Any, Dict, List

from loguru import logger

from turludock.helper_functions import check_if_remote_tag_exists, get_llvm_supported_versions, get_ubuntu_version
from turludock.yaml_load import load_cuda_config, load_cudnn_config


def check_if_cmake_version_exists(version: str) -> None:
    """Checks if a given CMake version exists in the remote repository.

    Args:
        version (str): The version of CMake to check.

    Raises:
        ValueError: If the given version of CMake is not found in the remote repository.
    """
    if not check_if_remote_tag_exists("https://github.com/Kitware/CMake.git", version):
        raise ValueError("CMake tag not found in remote. Check your configuration.")


def check_if_tmux_version_exists(version: str) -> None:
    """Checks if a given Tmux version exists in the remote repository.

    Args:
        version (str): The version of Tmux to check.

    Raises:
        ValueError: If the given version of Tmux is not found in the remote repository.
    """
    if not check_if_remote_tag_exists("https://github.com/tmux/tmux.git", version):
        raise ValueError("Tmux tag not found in remote. Check your configuration.")


def check_if_llvm_version_exists(version: str) -> None:
    """Checks if a given LLVM version exists in the remote repository.

    Args:
        version (str): The version of LLVM to check.

    Raises:
        ValueError: If the given version of LLVM is not found in the remote repository.
    """
    # Get supported LLVM versions
    supported_versions = get_llvm_supported_versions()

    # Check if version is supported
    version_is_supported = False
    for supported_version in supported_versions:
        try:
            if int(version) == int(supported_version):
                version_is_supported = True
        except Exception as e:
            logger.error(f"LLVM version not supported - check your configuration. Error: {e}")
            raise
    if not version_is_supported:
        raise ValueError(f"LLVM version {version} not supported. Supported are: {supported_versions}")


def check_against_known_list(config: Dict[str, Any], dict_key: str, supported_values: List[str]) -> None:
    """Checks if a given YAML key is in a list of supported values.

    In case of the extra packages, besides checking if we support the package, we also check if
    we support the version as well.

    Args:
        config (Dict[str, Any]): The configuration dictionary
        dict_key (str): The key in the configuration dictionary to check
        supported_values (list): The list of supported values for the given key

    Raises:
        ValueError: If the given value is not in the supported list of values
    """
    # In case the key points to a list
    if isinstance(config[dict_key], list):
        # Check each item of the config-list if it is a known configuration, e.g. a supported by us "package"
        for item in config[dict_key]:
            # If item is a dictionary we assume a version has been specified
            if isinstance(item, dict):
                if len(item.keys()) != 1:
                    raise ValueError(f"'{dict_key}: - {item}' cannot be a list. Specify only one version.")
                if "cmake" in item.keys():
                    package_name = "cmake"
                    check_if_cmake_version_exists(item[package_name])
                if "tmux" in item.keys():
                    package_name = "tmux"
                    check_if_tmux_version_exists(item[package_name])
                if "llvm" in item.keys():
                    package_name = "llvm"
                    check_if_llvm_version_exists(item[package_name])
            # Else no version has be specified and we just need to install "latest" available version
            else:
                package_name = item

            if package_name not in supported_values:
                raise ValueError(f"'{dict_key}: - {item}' not supported. Supported are {supported_values}")
    # Else if it is a single value
    else:
        if config[dict_key] not in supported_values:
            raise ValueError(f"'{dict_key}: {config[dict_key]}' not supported. Supported are {supported_values}")


def check_supported_ros_version(config: Dict[str, Any]) -> None:
    """Checks if a given ROS version is supported.

    Args:
        config (Dict[str, Any]): The configuration dictionary

    Raises:
        ValueError: If the ROS version is not supported
    """
    dict_key = "ros_version"
    supported_ros1_values = ["noetic"]
    supported_ros2_values = ["humble", "iron", "jazzy"]
    supported_values = supported_ros1_values + supported_ros2_values
    check_against_known_list(config, dict_key, supported_values)


def check_supported_gpu_drivers(config: Dict[str, Any]) -> None:
    """Checks if a given GPU driver is supported.

    Args:
        config (Dict[str, Any]): The configuration dictionary

    Raises:
        ValueError: If the given GPU drive is not supported
    """
    dict_key = "gpu_driver"
    supported_values = ["nvidia", "mesa"]
    check_against_known_list(config, dict_key, supported_values)


def check_extra_packages(config: Dict[str, Any]) -> None:
    """Checks the 'extra packages' YAML configuration.

    Args:
        config (Dict[str, Any]): The configuration dictionary

    Raises:
        ValueError: If the given GPU drive is not supported
    """
    dict_key = "extra_packages"
    # It can happen that no extra packages have been configured
    if dict_key not in config:
        logger.debug("No extra_packages have been configured in the .yaml file.")
        return
    supported_by_default = ["cmake"]
    supported_values = ["tmux", "llvm", "vscode", "conan", "meld", "cpplint"]
    supported_values += supported_by_default
    check_against_known_list(config, dict_key, supported_values)


def is_cuda_version_supported(cuda_version: str, ubuntu_version: str) -> bool:
    """Checks if the given CUDA version is supported for the given Ubuntu version.

    Args:
        cuda_version (str): CUDA version to check against the ubuntu_version
        ubuntu_version (str): The 'flat' Ubuntu version to check, e.g. 'ubuntu2204'

    Returns:
        bool: True if CUDA version is supported for the given Ubuntu version, False otherwise
    """
    cuda_config = load_cuda_config()
    if cuda_version in cuda_config:
        if ubuntu_version in cuda_config[cuda_version]:
            logger.debug(f"Found supported cuda-{cuda_version}-{ubuntu_version}")
            return True
        else:
            logger.error(f"Did not find supported ubuntu version '{ubuntu_version}' for cuda-{cuda_version}")
            supported_cuda_versions = list()
            for cuda_version in cuda_config:
                if ubuntu_version["flat"] in cuda_config[cuda_version]:
                    supported_cuda_versions.append(cuda_version)
            if len(supported_cuda_versions) == 0:
                logger.error(f"'cuda_version: {cuda_config['cuda_version']}' not supported at all!")
            else:
                logger.error(
                    f"cuda_version: {cuda_config['cuda_version']}' not supported. Supported are "
                    + f"{supported_cuda_versions} for Ubuntu {ubuntu_version['semantic']}"
                )
            return False
    else:
        logger.error(f"Did not find supported cuda version '{cuda_version}'")
        return False


def check_cuda_version(config: dict):
    """Check if the provided CUDA version is supported.

    Args:
        config (dict): The configuration dictionary

    Raises:
        ValueError: If the given CUDA version is not in the supported
    """
    ubuntu_version = get_ubuntu_version(config["ros_version"])
    if not is_cuda_version_supported(config["cuda_version"], ubuntu_version["flat"]):
        raise ValueError("CUDA version not supported.")


def is_cuda_cudnn_version_combination_supported(cuda_version: str, cudnn_version: str, ubuntu_version: str) -> bool:
    """Checks if the given cuDNN version is supported.

    A cuDNN version is always coupled to a CUDA version. Not all cuDNN/CUDA combinations are supported.
    And of course not all CUDA versions are supported by each Ubuntu version.
    So an Ubuntu version dictates which CUDA versions are supported and in turn each CUDA version
    dictates which cuDNN versions are supported.

    Args:
        cuda_version (str): the CUDA version to check
        cudnn_version (str): cuDNN version to check
        ubuntu_version (str): Ubuntu version to check

    Returns:
        bool: True if the CUDA and cuDNN version combination is supported for the given Ubuntu version, False otherwise
    """
    cudnn_config = load_cudnn_config()
    if cudnn_version in cudnn_config:
        if ubuntu_version in cudnn_config[cudnn_version]:
            if cuda_version in cudnn_config[cudnn_version][ubuntu_version]["cuda_version"]:
                return True
            else:
                logger.error(
                    f"Did not find supported cuDNN version '{cudnn_version}' "
                    + f"for cuda-{cuda_version} and Ubuntu '{ubuntu_version}'"
                )
                return False
        else:
            logger.error(f"Did not find supported Ubuntu version '{ubuntu_version}' for cuDNN {cudnn_version}")
            return False
    else:
        logger.error(f"Did not find supported cuDNN version '{cudnn_version}'")
        return False


def check_cudnn_version(config: dict):
    """Check if the provided cuDNN version is supported.

    Args:
        config (dict): The configuration dictionary

    Raises:
        ValueError: If the given cuDNN version is not in the supported
    """
    ubuntu_version = get_ubuntu_version(config["ros_version"])
    if not is_cuda_cudnn_version_combination_supported(
        config["cuda_version"], config["cudnn_version"], ubuntu_version["flat"]
    ):
        raise ValueError("cuDNN version not supported.")


def check_required_fields(config: dict):
    """Check if the configuration dictionary contains the required fields.

    Args:
        config (dict): The configuration dictionary

    Raises:
        ValueError: If the key 'ros_version' is missing in the YAML file
        ValueError: If the key 'gpu_driver' is missing in the YAML file
    """
    if "ros_version" not in config:
        raise ValueError("Please a ROS version, e.g. 'ros_version: noetic'")

    if "gpu_driver" not in config:
        raise ValueError("Please a GPU driver, e.g. 'gpu_driver: mesa'")


def check_nvidia_config(config: dict):
    """Check the CUDA/cuDNN configuration.

    Args:
        config (dict): The configuration dictionary

    Raises:
        ValueError: If the is a miss-configuration in the CUDA/cuDNN keys
    """
    try:
        # CUDNN only works with CUDA. So CUDA must be present
        if "cudnn_version" in config and "cuda_version" not in config:
            raise ValueError(
                "CUDNN version was configured, but not the CUDA version. " "Please configure also CUDA version."
            )

        # It doesn't make sense to install CUDA/CUDNN if the NVIDIA drivers are not present.
        if config["gpu_driver"] == "mesa":
            if "cuda_version" in config:
                logger.warning(
                    "CUDA has been configured although 'gpu_driver: mesa'. " "Did you mean 'gpu_driver: nvidia' ?"
                )
            if "cudnn_version" in config:
                logger.warning(
                    "CUDNN has been configured although 'gpu_driver: mesa'. " "Did you mean 'gpu_driver: nvidia' ?"
                )

        # Also given a warning in debug if nvidia has been selected but no CUDA or CUDNN
        if config["gpu_driver"] == "nvidia":
            if "cuda_version" not in config:
                logger.debug("Warning: 'cuda_version' is not set in the yaml configuration file")
            else:
                if "cudnn_version" not in config:
                    logger.debug(
                        "Warning: 'cudnn_version' is not set in the yaml configuration file although "
                        + f"'cuda_version': '{config['cuda_version']}'"
                    )

        # Check CUDA version
        if "cuda_version" in config:
            check_cuda_version(config)
            # Check CUDNN version
            if "cudnn_version" in config:
                check_cudnn_version(config)
    except ValueError as e:
        logger.error(f"NVIDIA configuration error: {e}")
        raise



import os
import tempfile
from typing import Any, Dict

import docker
from loguru import logger

import turludock.default_image_config as default_image_config
from turludock.build_progress import BuildProgress
from turludock.config_parser import check_dockerfile_config
from turludock.filesystem_operations import copy_resource, get_filename_from_path
from turludock.generate_dockerfile import generate_dockerfile
from turludock.yaml_load import load_yaml_file


def _generate_image_tag(yaml_config: Dict[str, Any]) -> str:
    """
    Generate the docker image tag based on the given configuration.

    Args:
        yaml_config (dict): The image configuration in yaml format.

    Returns:
        str: The generated docker image tag.
    """
    tag_name = "turlucode/"

    ros_version = yaml_config["ros_version"]
    tag_name += f"ros-{ros_version.lower()}"
    tag_name += f':{yaml_config["gpu_driver"].lower()}'

    if "cuda_version" in yaml_config:
        tag_name += f'-cuda{yaml_config["cuda_version"]}'
    if "cudnn_version" in yaml_config:
        tag_name += f'-cudnn{yaml_config["cudnn_version"]}'

    if "extra_packages" in yaml_config:
        for item in yaml_config["extra_packages"]:
            if isinstance(item, dict):
                package_name = next(iter(item), None)
            elif isinstance(item, str):
                package_name = item
            tag_name += f"-{package_name}"

    return tag_name


def build_image(docker_image_path: str, build_args: dict) -> None:
    """Build a Docker image using docker api

    Args:
        docker_image_path (str): The path to the Dockerfile to build.
        build_args (dict): The build arguments to use.
    """
    try:
        # Initialize the progress bar if needed
        if not build_args["verbose"]:
            build_progress = BuildProgress()

        # Connect to the Docker daemon
        client = docker.from_env()

        # Build the Docker image
        # Not using client.images.build so we can monitor the progress in real-time
        # See also: https://github.com/docker/docker-py/issues/376#issue-46825714
        response = client.api.build(
            path=docker_image_path,
            rm=True,  # Remove intermediate containers after a successful build
            tag=build_args["tag"],
            decode=True,  # The returned stream will be decoded into dicts on the fly
            nocache=build_args["no_cache"],  # Don't use the cache
        )

        # Process and print build logs in real-time
        try:
            for chunk in response:
                if "stream" in chunk:
                    if build_args["verbose"]:
                        print(chunk["stream"], end="", flush=True)
                    else:
                        build_progress.advance(chunk["stream"])
                elif "errorDetail" in chunk:
                    if build_args["verbose"]:
                        print(chunk["errorDetail"]["message"], end="", flush=True)
                    raise RuntimeError(f"Docker build error: {chunk['errorDetail']['message']}")
        except Exception as e:
            logger.error(f"client.api.build Error: {e}")
            raise

        if not build_args["verbose"]:
            build_progress.finish()

        # Print the ID of the built image
        image = client.images.get(build_args["tag"])
        print("")
        logger.info(f"Built image: '{build_args['tag']}' ({image.id})")
    except Exception as e:
        logger.error(f"Could not build image. Error: {e}")
        raise


def build_image_from_yaml_config(yaml_config: dict, build_args: dict) -> None:
    """Build a Docker image whose Dockerfile generation is based on the provided YAML config

    Args:
        yaml_config (dict): The YAML configuration for the auto-generation of the Dockerfile
        build_args (dict): The build arguments for 'docker build' command
    """
    # Check Dockerfile .yaml configuration
    check_dockerfile_config(yaml_config)

    # Generate Dockerfile based on configuration
    dockerfile = generate_dockerfile(yaml_config)

    if build_args["tag"] is None:
        build_args["tag"] = _generate_image_tag(yaml_config)

    # Create a temporary directory where we store the generated Dockerfile and its assets.
    # Important: when TemporaryDirectory() goes out of scope it deletes it.
    # So everything needs to happen within 'tempfile.TemporaryDirectory()'
    with tempfile.TemporaryDirectory() as temp_dir:
        # print(f"Temporary directory created at: {temp_dir}")

        # Store generated Dockerfile in tmp directory
        dockerfile_path = os.path.join(temp_dir, "Dockerfile")
        with open(dockerfile_path, "w", encoding="utf-8") as file:
            file.write(dockerfile)

        # Copy over Dockerfile assets
        copy_resource("turludock.assets.dockerfile_assets", "entrypoint_setup.sh", temp_dir)
        copy_resource("turludock.assets.dockerfile_assets", "terminator_config", temp_dir)

        # Build image
        build_image(temp_dir, build_args)


def build_pre_configured_image(config_name: str, build_args: dict) -> None:
    """Build an image given a provided by us configuration, a.k.a. pre-configuration

    All the pre-configuration are located in 'assets/default_image_configurations' and
    these are the ones we support basically.

    Args:
        config_name (str): The name of the pre-configured image to build
        build_args (dict): The build arguments for 'docker build' command
    """
    try:
        yaml_config = default_image_config.get_yaml_config(config_name)
        build_image_from_yaml_config(yaml_config, build_args)
    except Exception as e:
        logger.error(f"Could not build pre-configured image. {e}")
        raise


def build_custom_image(yaml_config_path: str, build_args: dict) -> None:
    """Build a Docker image based on a provided/custom YAML configuration

    Args:
        yaml_config_path (str): The path to the YAML configuration
        build_args (dict): The build arguments for the 'docker build' command
    """
    try:
        # Load .yaml file
        yaml_config = load_yaml_file(yaml_config_path)
        yaml_config.update({"filename": get_filename_from_path(yaml_config_path)})
        # Build custom-image
        build_image_from_yaml_config(yaml_config, build_args)
    except Exception:
        logger.error("Could not build custom-image")
        raise



import os

from loguru import logger
from termcolor import colored

import turludock.constants as constants
from turludock.config_parser import check_dockerfile_config, get_config_filename, print_configuration
from turludock.helper_functions import get_ubuntu_version, is_ros_version_supported, list_packaged_yaml_files
from turludock.yaml_load import load_cuda_config, load_cudnn_config, load_default_image_configuration


def _explain_default_image_config(yaml_filename: str) -> None:
    """Explains with one line in the terminal the current configuration

    Args:
        yaml_filename (str): The name of the YAML file configuration to explain.
    """
    # Load yaml configuration file
    yaml_config = load_default_image_configuration(yaml_filename)

    # Check if config is valid
    # TODO(ATA): Do we really need this check? It is slowing down the command
    check_dockerfile_config(yaml_config)

    # Print configuration to terminal
    print_configuration(yaml_config)


def _sort_file_list_based_on_release_date(config_files: list) -> list:
    """Short the list of config files based on their release date and then alphabetically

    Important!!! For this to work all presets need to start with "<ros_codename>_"!
    Otherwise this is not able to infer the release date and in turn sort those.

    Args:
        config_files (list): The file list

    Returns:
        list: The input list sorted
    """

    def get_ros_codename(string: str) -> str:
        """Get the ROS code name from the string using ROS_VERSION_RELEASE_DATE_MAP

        Args:
            string (str): Input string

        Returns:
            str: The ROS codename
        """
        for key, _ in constants.ROS_VERSION_RELEASE_DATE_MAP.items():
            if key in string:
                return key

    # Function to extract the sorting key
    def sort_key(string: str) -> tuple:
        """Provide the sorting key

        - Fist key is the release date since epoch
        - Second key is what ever is after "codename_" and sorted alphabetically

        Args:
            string (str): The filepath of the .yaml file

        Returns:
            tuple: The sorting tuple
        """
        # Split the string on the last part after the last '/'
        last_part = string.rsplit("/", 1)[-1]
        # Split the last part on the first '_'
        first_part, rest = last_part.split("_", 1)
        return (constants.ROS_VERSION_RELEASE_DATE_MAP[get_ros_codename(first_part)], rest)

    # Sort the list using the custom key
    sorted_list = sorted(config_files, key=sort_key)
    return sorted_list


def list_pre_configs() -> None:
    """List in the terminal available image pre-configurations as provided by our module."""
    print("")
    logger.info("Available pre-configurations:")
    yaml_config_files = list_packaged_yaml_files(os.path.join("assets", "default_image_configurations"))
    sorted_yaml_config_files = _sort_file_list_based_on_release_date(yaml_config_files)

    for yaml_config_full_path in sorted_yaml_config_files:
        yaml_filename: str = get_config_filename(yaml_config_full_path)
        _explain_default_image_config(yaml_filename)

    config_name: str = "CONFIG_NAME"
    cmd_example: str = colored(f"turludock build -e {config_name}", attrs=["bold"])
    print("")
    print(f"> You can directly build a default docker image with: {cmd_example}")


def list_supported_ros_versions() -> None:
    """List in the terminal all supported ROS distributions."""
    print("")
    ros1 = [version for version, ros_version in constants.ROS_VERSION_MAP.items() if ros_version == 1]
    ros2 = [version for version, ros_version in constants.ROS_VERSION_MAP.items() if ros_version == 2]
    logger.info("Supported ROS Versions")
    logger.info(f"ROS 1: {ros1}")
    logger.info(f"ROS 2: {ros2}")


def list_cuda_support(ros_codename: str) -> None:
    """Lists all supported CUDA/cuDNN versions for a given ROS distribution.

    Args:
        ros_codename (str): ROS codename
    """
    ros_codename = ros_codename.lower()
    try:
        # Check if we even support this ROS version
        if not is_ros_version_supported(ros_codename):
            return

        # Load CUDA and cuDNN configs
        cuda_config = load_cuda_config()
        cudnn_config = load_cudnn_config()

        # Find which ones we support
        found_supported = False
        ubuntu_version = get_ubuntu_version(ros_codename)
        print("")
        logger.info(f"Supported versions for Ubuntu {ubuntu_version['semantic']} ({ros_codename}):")
        for cuda_version in cuda_config:
            if ubuntu_version["flat"] in cuda_config[cuda_version]:
                logger.info(f"*CUDA: {cuda_version}")
                found_supported = True
                for cudnn_version in cudnn_config:
                    if ubuntu_version["flat"] in cudnn_config[cudnn_version]:
                        if cudnn_config[cudnn_version][ubuntu_version["flat"]]:
                            logger.info(f"*CUDA: {cuda_version} | cuDNN: {cudnn_version}")
        if not found_supported:
            logger.warning(f"No supported CUDA/cuDNN version for ROS {ros_codename.capitalize()} at this point.")
    except Exception:
        logger.error("Could not run 'list-cuda-support' command. Exit.")
        raise



import os

from loguru import logger

import turludock.default_image_config as default_image_config
from turludock.config_parser import check_dockerfile_config
from turludock.filesystem_operations import copy_resource, get_filename_from_path
from turludock.generate_dockerfile import generate_dockerfile
from turludock.yaml_load import load_yaml_file


def _populate_build_folder(yaml_config: dict, dir_path: str) -> None:
    """Populate the provided directory with the generated Dockerfile and its assets

    Args:
        yaml_config (dict): The configuration dictionary
        dir_path (str): The path of the directory where to populate the files
    """
    # Check Dockerfile .yaml configuration
    check_dockerfile_config(yaml_config)

    # Generate Dockerfile based on configuration
    dockerfile = generate_dockerfile(yaml_config)

    # Store generated Dockerfile in directory
    dockerfile_path = os.path.join(dir_path, "Dockerfile")
    with open(dockerfile_path, "w", encoding="utf-8") as file:
        file.write(dockerfile)
    logger.debug(f"Successfully copied generated Dockerfile in '{dir_path}'")

    # Copy over Dockerfile assets
    copy_resource("turludock.assets.dockerfile_assets", "entrypoint_setup.sh", dir_path)
    copy_resource("turludock.assets.dockerfile_assets", "terminator_config", dir_path)


def check_if_directory_path_is_valid(path: str) -> None:
    """Check if we have a valid path to a directory.

    Args:
        path (str): The directory path to check.

    Raises:
        ValueError: If the path is not a valid directory
        ValueError: If the directory does not exist or we do not have write access.
    """
    if not os.path.isdir(path):
        raise ValueError(f"The path '{path}' is not a valid directory.\n")
    if not os.access(path, os.W_OK):
        raise ValueError(f"We do not have write access to '{path}'\n")


def generate_from_pre_config(config_name: str, dir_path: str) -> None:
    """Populate the build folder with the Dockerfile and its assets using provided pre-configurations.

    See 'assets/default_image_configurations' for the list of supported pre-configurations.

    Args:
        config_name (str): The name of the pre-defined configuration to use.
        dir_path (str): The path to the directory where to store the generated Dockerfile and its assets

    Raises:
        Exception: If there is a problem populating the folder.
    """
    try:
        check_if_directory_path_is_valid(dir_path)
        yaml_config = default_image_config.get_yaml_config(config_name)
        _populate_build_folder(yaml_config, dir_path)

        print("")
        logger.info(f"Populated folder: '{dir_path}'")
    except Exception:
        logger.error(f"Could not populate build folder '{dir_path}'.")
        raise


def generate_from_user_config(yaml_config_path: str, dir_path: str):
    """Populate the build folder with the Dockerfile and its assets using the custom YAML configuration.

    Args:
        yaml_config_path (str): The path to the custom YAML configuration.
        dir_path (str): The path to the directory where to store the generated Dockerfile and its assets

    Raises:
        Exception: If there is a problem populating the folder.
    """
    try:
        check_if_directory_path_is_valid(dir_path)
        yaml_config = load_yaml_file(yaml_config_path)
        yaml_config.update({"filename": get_filename_from_path(yaml_config_path)})
        _populate_build_folder(yaml_config, dir_path)

        print("")
        logger.info(f"Populated folder: '{dir_path}'")
    except Exception:
        logger.error(f"Could not populate build folder '{dir_path}'.")
        raise
